from rstt.ranking import Standing
from rstt.ranking.ranking import Ranking, get_disamb
from rstt.ranking.datamodel import KeyModel, GaussianModel
from rstt.ranking.rating import GlickoRating
from rstt.ranking.inferer import Glicko, Elo, PlayerLevel
from rstt.ranking.observer import  GameByGame, BatchGame, KeyChecker


class BTRanking(Ranking):
    def __init__(self, name: str='', players=None):
        super().__init__(name=name,
                        standing=Standing(),
                        datamodel=KeyModel(factory=lambda x: x.level()),
                        backend=PlayerLevel(),
                        handler=KeyChecker(),
                        players=players)
        
    def forward(self, *args, **kwargs):
        # no args, kwargs
        self.handler.handle_observations(infer=self.backend,
                    datamodel=self.datamodel)


class BasicElo(Ranking):
    def __init__(self, name: str='', default: float=1500.0, k: float=20.0, lc: float=400.0, players=None):
        super().__init__(name=name,
                        standing=Standing(default=default),
                        datamodel=KeyModel(default=default),
                        backend=Elo(k=k, lc=lc),
                        handler=GameByGame(),
                        players=players)
    
    def predict(self, game):
        _, teams_as_ratings, _, _ = self.handler.match_formating(datamodel=self.datamodel, game=game)
        r1 = teams_as_ratings[0][0]
        r2 = teams_as_ratings[1][0]
        return self.backend.expectedScore(rating1=r1, rating2=r2)


class BasicGlicko(Ranking):
    def __init__(self, name: str='', handler=BatchGame(), mu: float=1500.0, sigma: float=350.0, players=None):
        super().__init__(name= name,
                        standing=Standing(),
                        datamodel=GaussianModel(default=GlickoRating(mu, sigma)),
                        backend=Glicko(),
                        handler=handler,
                        players=players)

    @get_disamb
    def __step1(self):
        # TODO: check which player iterator to use
        for player in self:
            rating = self.datamodel.get(player)
            rating.sigma = self.backend.prePeriod_RD(rating)

    def forward(self, *args, **kwargs):
        self.__step1()
        self.handler.handle_observations(infer=self.backend, datamodel=self.datamodel, *args, **kwargs)
    
    def predict(self, game):
        _, teams_as_ratings, _, _ = GameByGame().match_formating(datamodel=self.datamodel, game=game)
        r1 = teams_as_ratings[0][0]
        r2 = teams_as_ratings[1][0]
        return self.backend.expectedScore(rating1=r1, rating2=r2)


class BasicOS(Ranking):
    def __init__(self, name: str='', model=None, players=None):
        super().__init__(name=name,
                        standing=Standing(),
                        datamodel=GaussianModel(factory=lambda x: model.rating(name=x.name)),
                        backend=model,
                        handler=GameByGame(),
                        players=players)
    
    def predict(self, game) -> float:
        _, teams_as_ratings, _, _ = self.handler.match_formating(datamodel=self.datamodel, game=game)
        return self.backend.predict_win(teams_as_ratings)[0]
    
    def quality(self, game) -> float:
        _, teams_as_ratings, _, _ = self.handler.match_formating(datamodel=self.datamodel, game=game)
        return self.backend.predict_draw(teams_as_ratings)


from typing import List, Callable
from typeguard import typechecked

from rstt import Duel
import rstt.utils.functions as uf

import random



'''

    TODO:
    - Extend match to Many-Versus-Many match
    - Extend match to Free-for-all
    - LEVEL_MIXTURES: define differents ways to mix levels in a teams, sum/avg/median/ and set a parameters to tune it solvers
    - Create const value for standard score (maybe enum types) i.e Score.win := [1,0]| Score.lose := [0,1]/ Score.draw := [0.5, 0.5]
    - Work on Score type
    - Add predict() to solvers (and to the stypes.Solver Protocol ?)

'''

type Score = list[float]

WIN = [1.0, 0.0]
LOSE = [0.0, 1.0]
DRAW = [0.5, 0.5]

        
class BetterWin():
    def __init__(self, with_draw: bool=False):
        self.with_draw = with_draw
        self.win = WIN
        self.lose = LOSE
        self.draw = DRAW
        
    def solve(self, duel: Duel, *args, **kwars) -> None:
        level1, level2 = duel.player1().level(), duel.player2().level()
        if level1 > level2:
            score = self.win
        elif level1 < level2:
            score = self.lose
        elif self.with_draw:
            score = self.draw
        else:
            score = self.win
        duel._Match__set_result(result=score)
                

class ScoreProb:
    @typechecked
    def __init__(self, scores: List[Score], func: Callable[[Duel], Score]):
        self.scores = scores
        self.probabilities = func
    
    def solve(self, duel: Duel, *args, **kwars) -> None:
        score = random.choices(population=self.scores, 
                               weights=self.probabilities(duel),
                               k=1)[0]
        duel._Match__set_result(score)


class WeightedScore(ScoreProb):
    def __init__(self, scores: List[Score], weights: List[float]):
        if len(scores) != len(weights):
            msg = f"length of scores ({len(scores)}) does not match length of weights ({len(weights)})"
            raise ValueError(msg)
        super().__init__(scores=scores, func= lambda x: weights)


class CoinFlip(WeightedScore):
    def __init__(self):
        super().__init__(scores=[WIN, LOSE], weights=[0.5, 0.5])
        

class BradleyTerry(ScoreProb):
    def __init__(self):
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)
    
    def __probabilities(self, duel=Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.bradleyterry(level1, level2)
        return [prob, 1-prob]

class LogSolver(ScoreProb):
    def __init__(self, base: float=10, lc: float=400):
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)
        self.base = base
        self.lc = lc
    
    def __probabilities(self, duel: Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.logistic_elo(base=self.base, diff=level1-level2, constant=self.lc)
        return [prob, 1-prob]




from typing import Union, List, Set, Dict, Optional
from typeguard import typechecked
import abc

from rstt import Player, Duel, BetterWin
from rstt.stypes import Solver, Achievement
from rstt.ranking.ranking import Ranking
import rstt.utils.utils as uu

from collections import defaultdict


def playersToDuel(players: List[Player]) -> List[Duel]:
    return [Duel(players[2*i], players[2*i+1]) for i in range(len(players)//2)]    


class Competition(metaclass=abc.ABCMeta):
    @typechecked
    def __init__(self, name: str,
                 seeding: Ranking,
                 solver: Solver=BetterWin(),
                 cashprize: Optional[Dict[int, float]]=None):
        # a name
        self.name = name
        
        # 'settings'
        self.participants = []
        self.seeding = seeding
        self.solver = solver
        self.cashprize = defaultdict(lambda: 0)
        if cashprize:
            self.cashprize.update(cashprize)
        
        # result related variable
        self.played_matches = []
        self.standing = {}
        
        # control variable
        self.__started = False
        self.__finished = False
        self.__closed = False

    # --- getter --- #
    def started(self):
        return self.__started
    
    def live(self):
        return self.__started and not self.__finished
    
    def over(self):
        return self.__closed
    
    def games(self, by_rounds=False):
        return self.played_matches if by_rounds else uu.flatten(self.played_matches)
    
    def top(self, place: Optional[int]=None) -> Union[Dict[int, List[Player]], List[Player]]:
        if place:
            return [key for key, value in self.standing.items() if value == place]
        else:
            return {v: [key for key, value in self.standing.items() if value == place] for v in self.standing.values()}

    @typechecked
    def registration(self, players: Union[Player, List[Player], Set[Player]]):
        if not self.__started:
            playerset = set(self.participants)
            playerset.update(players)
            self.participants = list(playerset)

    def run(self):
        if not self.__started:
            self.start()
            self.play()
            self.trophies()

    def start(self):
        if not self.__started:
            self.seeding = self.seeding.fit(self.participants)
            self.initialise()
            self.__started = True

    def play(self):
        while not self.__finished:
            current_round = self.generate_games()
            results = self.play_games(current_round)
            self.__finished = self.edit(results)

    @typechecked
    def play_games(self, games: List[Duel]):
        played = []
        for game in games:
            self.solver.solve(game)
            played.append(game)
        self.played_matches.append(played)
        return played

    def trophies(self):
        for player in self.participants:
            try: 
                result = Achievement(self.name, self.standing[player], self.cashprize[self.standing[player]])
                player.collect(result)
            except AttributeError:
                continue
        self.__closed= True

    def initialise(self):
        pass

    @abc.abstractmethod
    def generate_games(self):
        pass

    @abc.abstractmethod
    def edit(self, games: List[Duel]):
        return True
    


from dataclasses import dataclass
from typing import Any, Protocol, runtime_checkable

from rstt import Player, Match


# TODO: define 'Event', 'Tournament', 'MatchMaking'




@dataclass
class Achievement:
    event_name: str
    place: int
    prize: float
    # ??? categorie (world event, domestic league, playoffs etc.)
    # ??? event type (SEB, Snake, MM etc)
    # ??? teamates
    
    
@runtime_checkable
class Solver(Protocol):
    def solve(self, match: Match, *agrs, **kwargs) -> None:
        ...


@runtime_checkable
class Inference(Protocol):
    # NOTE: name source: https://en.wikipedia.org/wiki/Statistical_inference
    def rate(self, *args, **kwargs) -> Any:
        ...


@runtime_checkable        
class RatingSystem(Protocol):
    def get(self, key: Player) -> Any:
        ...
        
    def set(self, key: Player, rating: Any) -> None:
        ...
        
    def ordinal(self, rating: Any) -> float:
        ...


@runtime_checkable
class Observer(Protocol):
    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs) -> None:
        ...




import random



# -------------------- #
# --- Player cfg ----- #
# -------------------- #

PLAYER_GAUSSIAN_MU = 1500
PLAYER_GAUSSIAN_SIGMA = 500

PLAYER_DIST = random.gauss
PLAYER_DIST_ARGS = {'mu': PLAYER_GAUSSIAN_MU,
             'sigma': PLAYER_GAUSSIAN_SIGMA}


from typing import Any, Callable

from dataclasses import dataclass
from collections import defaultdict

import copy

from rstt.player import Player
    
    
class keydefaultdict(dict[Player, Any]):
    ''' a defaultdict accpeting lambda x: func(x)
    source:
    Minor modification from the answer of Paulo Costa in: 
    https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory
    '''
    def __init__(self, default_factory: Callable[[Player], Any]):
        super().__init__()
        self.default_factory = default_factory

    def __missing__(self, key: Player) -> Any:
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class KeyModel:
    def __init__(self, default=None, template=None, factory=None, *args, **kwargs):
        self.ratings = self.__init_ratings(default, template, factory, *args, **kwargs)
        self.__rtype = self.__get_rating_type()
        self.__default = self.__get_default_rating()

    # --- setter --- #
    def set(self, key, rating):
        # TODO: test rating type before assignement and thorw TypeError
        self.ratings[key] = rating

    # --- getter --- #
    def get(self, key):
        # QUEST: __getitem__ ?
        return self.ratings[key]

    def items(self):
        # ??? ambiguous behaviour items() sugest a collection of (key, rating)
        return self.ratings
    
    def keys(self):
        return self.ratings.keys()

    # --- general purpose methords --- #
    def rtype(self):
        return self.__rtype
        
    def default(self):
        return self.__default

    def ordinal(self, rating) -> float:
        # REQ: Should not support arg type Player. This is the job of a Ranking/Standing 
        # NOTE: name source: https://fr.wikipedia.org/wiki/Nombre_ordinal
        '''convert a rating object into a floating value'''
        return float(rating)

    def tiebreaker(self, rating):
        try:
            return list(rating)
        except:
            return 0

    # --- internal mechanism --- #     
    def __init_ratings(self, default, template, factory, *args, **kwargs):
        ''' rating initialization
        
        return: defaultdict {key: rating}
            where key are the one contained in the standing
            and rating are object used to compute the associated value. 
        
        REQ:
            - each key as its own rating
            - user can provide a rating object (a) or a type to use as model (b).
            - these to option are imcompatible and an error need to be raised
            - match the default value with the self.__rate_model(rating) method
            (a) We use deepcopy to create a new rating instance with equal value.
            (b) We use a Constructor with params to generate a new rating object.
        
        TODO:
            - check functool.partial
        '''
        ratings = {}
        if default and not template and not factory:
            ratings = self.__default_ratings(value=default)
        elif template and not default and not factory:
            ratings = self.__template_ratings(template, *args, **kwargs)
        elif factory and not default and not template:
            ratings = self.__factory_ratings(factory, *args, **kwargs)
        else:
            # TODO: write incompatible params error msg
            msg = f""
            raise ValueError(msg)
        return ratings
    
    def __default_ratings(self, value):
        return defaultdict(lambda: copy.deepcopy(value))
    
    def __template_ratings(self, template, *args, **kwargs):
        return defaultdict(lambda: template(*args, **kwargs))
    
    def __factory_ratings(self, func: Callable, *args, **kwargs):
        return keydefaultdict(default_factory=func)

    def __get_rating_type(self):
        dummy = Player('dummy', 0.0)
        rtype = type(self.ratings[dummy])
        del self.ratings[dummy]
        return rtype

    def __get_default_rating(self):
        dummy = Player('dummy', 0.0)
        self.ratings[dummy]
        return self.ratings.pop(dummy)
    
    # --- magic methods --- #
    def __delitem__(self, key: Player):
        del self.ratings[key]
    
    # ??? __setitem__
    # ??? __getitem__
    # ??? __contain__
        

class GaussianModel(KeyModel):
    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        
    def ordinal(self, rating):
        return rating.mu - 2*rating.sigma
        
    def tiebreaker(self, rating):
        return [rating.mu, rating.sigma]




from typeguard import typechecked
from typing import Any, Union, List, Callable

from rstt import Player
from rstt.ranking import Standing
from rstt.stypes import Inference, RatingSystem, Observer


def set_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return set_action
    return wrapper_set

def get_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return func(self, *args, **kwars)
    return wrapper_get
    
def get_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return func(self, *args, **kwars)
    return wrapper_get
    
def set_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return set_action
    return wrapper_set


class Ranking():
    '''   
    NOTE:
        - The Ranking class has two dict[Player, ...], (RatingSystem, Standing) - which induce redundancy and ambiguity.
        The 'Container Equivalence' and 'Rank Disambiguity' (SEE GLOSSARY) are two properties that needs to be enforced
        to guarantee a well defined behaviour for the Ranking class. Hopefully it is also sufficient.
        - The RatingSystem act like a defaultdict. Get operations can induce set operations.
        This is tricky as a player could be inserted in the ranking using 'read' methods of the interface.
        
    GLOSSARY:
        - Container Equivalence (Union == Intersection):
        (key in self.datamodel.ratings) <=> (key in self.standing).
        In the code we refer to 'equivalence'
        - Rank Disambiguity (point '=' rating):
        self.datamodel.ordinal(key) == self.standing.value(key) for all keys.
        In the code we refer to 'disambiguity'

    # !!!
        - self.__disambiguity == True && self._Standing__sorted == False
        This means that the players have the correct points but are not correctly ranked.
        It can happen when 'self.standing._keep_sorted_ == False' which is triggered by the user outside of the Ranking methods.
    
    # ??? 
        - Can this state be reached from the intended Ranking usage ?
        How to detect it and what to do ?
    '''
    @typechecked
    def __init__(self, name: str = '',
                 standing: Standing = Standing(),
                 datamodel: RatingSystem = None, 
                 backend: Inference = None,
                 handler: Observer = None,
                 players: Union[None, List[Player]] = None):
        
        # name/identifier - usefull for plot
        self.name = name
        
        # fundamental notion of the Ranking Class
        self.standing = standing
        self.backend = backend
        self.datamodel = datamodel
        self.handler = handler

        # state control variable
        self.__equivalence = True
        self.__disambiguity = True

        # protocol control variable
        self.__maintain_equivalence = True
        self.__maintain_disambiguity = True
        
        if players:
            self.add(keys=players)
        
    # --- Containers standard methods --- #
    @set_disamb
    @set_equi
    @typechecked
    def add(self, keys: List[Player]):
        # turn maintainance for optimization
        should_maintain = self.__maintain_equivalence
        self.__maintain_equivalence = False
        
        # perform iteratively addition
        for key in keys:
            self.__add(key)
        
        # restaure Ranking status
        self.__maintain_equivalence = should_maintain
             
    def __add(self, key: Player):
        if key in self:
            msg = f'Can not add a key already present in the Ranking, {key}'
            raise ValueError(msg)
        # default dict get operator for missing key
        self.datamodel.get(key)
        
        # self.datamodel do not match self.standing
        self.__equivalence = False
    
    # --- magic methods --- #
    def __getitem__(self, *args, **kwargs) -> Union[Player, List[Player], int, List[int]]:
        ''' get item based on a rank or a key
        
        NOBUG:
            - sorting is handled by the Standing itself
            - so is typechecked
        '''
        return self.standing.__getitem__(*args, **kwargs)
    
    def __delitem__(self, key: Player):
        ''' delete element from the Ranking
        
        REQ:
            - element needs to be remove both from the standing and the RatingSystem
            
        NOBUG:
            - del standing[key] is typechecked. CALL MUST BE BEFORE RatingSystem
        
        # ???:
            - could del succeed on standing but fail on RatingSystem.
            This would otentialy lead to an invalid ranking state because __delitem__ is not decorated.
            A ranking invalid state can exactly be the reason why this scenario happens.
            What is the best approach to this situation ?
        '''
        del self.standing[key]
        del self.datamodel[key]
            
    def __contains__(self, key: Player):
        '''
        NOTE:
            - it does match standing behavior as specified but
            depending on the choices, 'p in self' can be slower than
            'p in self.standing' / 'p in self.datamodel'
        '''
        return key in self.standing
    
    def __len__(self):
        return len(self.standing)
    
    def __iter__(self):
        return self.standing.__iter__()
                
    # --- getter --- #
    def rank(self, player: Player) -> int:
        return self[player]
        
    def ranks(self, players: List[Player]) -> List[int]:
        return [self.rank(player) for player in players]
        
    def rating(self, player: Player) -> Any:
        """Get method for rating

        Rating object is the internal model associated to a key.
        Ratings are used to automaticly compute values for the sorting feature of a Standing.

        Parameters
        ----------
        player : Player
            A key in the Ranking

        Returns
        -------
        _type_
            The associated model to the provided key. The type is defined by Ranking.RatingSystem.rtype
            
        Raises
        ------
        KeyError
        """
        if player in self: # NOBUG RatingSystem is a defaultdict
            return self.datamodel.get(player)
        else:
            msg = f"{player} is not present in {self.standing}"
            raise KeyError

    def ratings(self) -> List[Any]:
        """Get method for all ratings

        Returns
        -------
        list[_type_]
            A list of all rating object present in the Ranking, in order of the Standing.
        """
        return [self.rating(player) for player in self]
    
    def players(self) -> List[Player]:
        """Get method of all keys

        Alias for Ranking.standing.keys()

        Returns
        -------
        List[Player]
            A list of all player in descending order of their associated values.
        """
        return self.standing.keys()

    def point(self, player: Player) -> float:
        """Get the point associated to a key

        Alias for Ranking.standing.value(player)

        Returns
        -------
        float
            the associated value.
        """
        return self.standing.value(player)

    def points(self) -> List[float]:
        """Get method of all values

        Alias for Ranking.standing.values()

        Returns
        -------
        List[float]
            A list of all associated values in descending order.
        """
        return self.standing.values()

    def status(self):
        return  {'equivalence': self.__equivalence,
        'disambiguity': self.__disambiguity,
        'maintain_equivalence': self.__maintain_equivalence,
        'm_disambanbiguity': self.__maintain_disambiguity}
        
    # ??? items() -> List[(rank, player, ratings)]
    # ??? item(key) -> (rank, player, ratings)
    # --- setter --- #
    @set_disamb
    @set_equi
    def set_rating(self, key: Player, rating: Any):
        """A method to assign a rating to a Player

        
        The Ranking delegate this task to a 'RatingSystem' instance stored as attribute 'rankink.datamodel'.
        The RatingSystem define what rating type is accepted and wether a set operation is authorized for the provided key.
        
        Parameters
        ----------
        key : Player
            A Player
        rating : Any
            A rating object associated to the key
        """
        self.datamodel.set(key, rating)
        self.__equivalence = False
        
    # ??? remove()
    # ??? pop()
            
    # --- general purpose methods --- #
    @get_disamb
    @get_equi   
    def plot(self):
        self.standing.plot(standing_name=self.name)
    
    @set_disamb
    @set_equi
    def update(self, *args, **kwargs):
        """Update the Ratings

        # TODO: explain what happen here. GL&HF

        Parameters
        ----------
        observations : _type_
            _description_
        """
        
        self.forward(*args, **kwargs)
      
        # BUG: How do we know if the ranking state changed ?
        # HACK: assume we did change everything
        self.__disambiguity = False
        self.__equivalence = False
    
    def forward(self, *args, **kwargs):
        '''
        BUG:
            - while using BasicElo
        TypeError: GameByGame.handle_observations() got multiple values for argument 'infer'
        This was fixed 'user side' by giving a param 'games' to GameByGame.handle_observations() and explicilty calling
        Ranking.update(games=[])
        But it should not be necessary to make an explicit arg assignement. 
        And the source of the error never identified.
        
        It happens in the wrapper. an explanation could be related to the self parameters
        
        '''
        self.handler.handle_observations(infer=self.backend,
                            datamodel=self.datamodel,
                            *args, **kwargs)

    @get_disamb
    @get_equi
    def rerank(self, permutation: List[int], name: str = None, direct: bool = True):
        # TODO: write doc
        # check permutation validity
        if not (set(permutation) == set(i for i,_ in enumerate(self))):
            # TODO: write good msg error
            msg = ''
            raise ValueError(msg)
        
        # TODO: deal with direct=False
        
        pairs = []
        for current_rank, future_rank in enumerate(permutation):
            player = self[current_rank]
            ratings = self.datamodel.get(self[future_rank])
            pairs.append((player, ratings))
        for p, r in pairs:
            self.datamodel.set(p, r)
    
    @get_disamb
    @get_equi
    def fit(self, players: List[Player], name: str = ''):
        seeding= Standing()
        points = []
        for player in players:
            if player in self:
                points.append(self.point(player))
            else:
                points.append(self.standing._default_)
        seeding.add(players, points)
        return seeding
    
    # --- internal mechanism --- #
    def __ContainerEquivalence(self):
        ''' property checker
        
        '''
        # get keys
        standing_keys = set(self.standing.keys())
        RatingSystem_keys = set(self.datamodel.keys())
        
        if standing_keys == RatingSystem_keys:
            self.__equivalence = True
        elif standing_keys <= RatingSystem_keys: # a <= b means 'a.issubset(b)
            not_ranked_players = list(RatingSystem_keys - standing_keys) # a - b means a.difference(b)
            new_points = []
            for player in not_ranked_players:
                # NOBUG: player is in the RatingSystem. 'get()' is safe to perform.
                new_points.append(self.datamodel.ordinal(self.datamodel.get(player)))
            # NOBUG: no ambiguity is introduce this way
            self.standing.add(keys=not_ranked_players, values=new_points)
            self.__equivalence = True
        else:
            # TODO: write a good error message
            msg = f''
            raise RuntimeError(msg)
        
    def __RankDisambiguity(self):
        for player in self.standing:
            # TODO: performance check (a) assign only if needed (b) assign always
            # ??? (a / b) as user options ?
            rating = self.datamodel.get(player)
            point = self.datamodel.ordinal(rating)
            if self.standing.value(player) != point:
                self.standing[player] = point
        
        self.__disambiguity = True




from typing import List, Dict
from typeguard import typechecked


from rstt.scheduler.tournament.competition import Competition, playersToDuel
from rstt.ranking.ranking import Ranking
from rstt import Duel, BetterWin
from rstt.stypes import Solver

import rstt.utils.matching as um
import rstt.utils.utils as uu

import math

def balanced_tree(rounds):
    matches = [1, 2]
    for round in range(2, rounds+1):
        check_sum = pow(2, round)+1
        new_matches = []
        for i in matches:
            new_matches.append([i, check_sum-i])
        matches = uu.flatten(new_matches)

    return matches


class SingleEliminationBracket(Competition):
    def __init__(self, name: str,
                seeding: Ranking,
                solver: Solver = BetterWin(),
                cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)
        self.players_left = []

    def initialise(self):
        if not uu.power_of_two(len(self.participants)):
            msg = (f'{type(self)} '
                   'needs a power of two as number of participants '
                   '(2,4,8,16,...)'
                   f', given {len(self.participants)}')
            raise ValueError(msg)

        nb_rounds = int(math.log(len(self.participants), 2))
        self.players_left = self.seeding[[i-1 for i in balanced_tree(nb_rounds)]]

    def generate_games(self):
        games = []
        for i in range(0, len(self.players_left)//2):
            p1 = self.players_left[2*i]
            p2 = self.players_left[2*i+1]
            match = Duel(p1, p2)
            games.append(match)
        return games

    def edit(self, games: List[Duel]):
        next = []
        for game in games:
            next.append(game.winner())
            self.standing[game.loser()] = len(self.players_left)
        self.players_left = next

        if len(self.players_left) == 1:
            self.standing[self.players_left[0]] = 1
            finished = True
        else:
            finished = False

        return finished


class DoubleEliminationBracket(Competition):
    def __init__(self, name: str,
                seeding: Ranking,
                solver: Solver = BetterWin,
                # TODO: add policy params
                cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)
        
        self.upper = SingleEliminationBracket(name+'_UpperBracket',
                                              seeding, solver, cashprize=None)
        self.lower = [] # List[Player]
        self.injector = [] # List[List[Player]]

        # NOTE: this is a um.riffle_shuffle for List[Player]
        #self.policy = lambda list1, list2: [e for e1, e2 in zip(list1, list2) for e in [e1, e2]]
        self.policy = um.riffle_shuffle

    def initialise(self):
        # play upper bracket as a 'SEB'
        self.upper.registration(self.participants)
        self.upper.start()
        self.upper.play()
        
        self.lower = [[game.loser() for game in r] for r in self.upper.games(by_rounds=True)]
        self.lower.append(self.upper.top(place=1))

    def generate_games(self):
        # lower bracket games
        if len(self.lower[0]) != len(self.lower[1]):
            
            games = playersToDuel(self.lower.pop(0))
        # injector games
        elif self.injector:
            lower = self.lower.pop(0)
            injector = self.lower.pop(0)
            games = playersToDuel(self.policy(lower, injector))
        return games

    def edit(self, games: List[Duel]):
        self.standing.update({
            game.loser(): len(self.participants)-len(self.standing)
            for game in games})
        self.lower.insert(0, [game.winner() for game in games])

        if not self.lower: # empty when the grand final is played
            self.standing[games[0].winner()] = 1
            self.standing[games[0].loser()] = 2
            finished = True
        else:
            finished = False
        return finished


from typing import List, Union, Optional
from typeguard import typechecked

import abc

from rstt.stypes import Achievement
from rstt import Player, Match

import random


class HistoryPlayer(Player):
    @typechecked
    def __init__(self, name: Optional[str]=None, level: Optional[float]=None) -> None:
        super().__init__(name=name, level=level)
        self.__achievements = []
        self.__games = []
    
    # --- getter --- #
    def achievements(self) -> List[Achievement]:
        return self.__achievements
    
    def earnings(self) -> float:
        return sum([achievement.prize for achievement in self.__achievements]) 
    
    def games(self) -> List[Match]:
        return self.__games   
    
    # --- setter --- #
    @typechecked
    def collect(self, achievement: Union[Achievement, List[Achievement]]):
        if isinstance(achievement, Achievement):
            achievements = [achievement]
        
        previous_event = [past_event.event_name for past_event in self.__achievements]
        for achievement in achievements:
            if achievement not in previous_event:
                self.__achievements.append(achievement)
            else: 
                msg=f"Can not collect {achievement}. {self} already participated in an event called {achievement.event_name}"
                raise ValueError(msg)

    def reset(self) -> None:
        self.__achievements = []
        self.__games = []
        
    # --- internal mechanism --- #
    @typechecked
    def add_game(self, match: Match) -> None:
            self.__games.append(match)
            
            
class PlayerTVS(Player, metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)
        self.__current_level = self._Player__level
        self.__level_history = []
        self.__params = kwars # level evolution parameters
        
    # --- getter --- #
    def level_history(self) -> List[float]:
        return self.__level_history
    
    def original_level(self) -> float:
        return self._Player__level
    
    def params(self):
        return self.__params
    
    # --- override --- #
    def level(self):
        return self.__current_level
    
    # --- internal mechanism --- #
    @abc.abstractmethod
    def update_level(self, *args, **kwars) -> None:
        '''change the self.__current_level value'''


class GaussianPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: str, mu: float, sigma: float):
        # pass mu as level to Player, and mu/sigam as params to PlayerTVS
        super().__init__(name, mu, mu=mu, sigma=sigma)
        self.__mu = mu
        self.__sigma = sigma
        
    def update_level(self, *args, **kwars) -> None:
        self._PlayerTVS__current_level = random.gauss(self.__mu, self.__sigma)
    
    
        
        
    
    


from typing import  List, Dict, Tuple, Any
from typeguard import typechecked

from rstt.player import Player

import math
import copy


class PlayerLevel:
    @typechecked
    def rate(self, player: Player, *args, **kwars) -> Dict[Player, float]:
        return {player: player.level()} # FIXME: standardize rate() output

    
class Elo:
    def __init__(self, k: float = 20.0, lc: float = 400.0):
        self.LC = lc
        self.K = k
        # TODO self.distribution = dist & change expectedScore

    @typechecked
    def rate(self, groups: List[List[float]], scores: List[float], *args, **kwars) -> List[List[float]]:        
        # NOTE: groups: [[winner_elo][loser_elo]], scores [[1.0][0.0]]
        
        # unpack args
        r1, r2 = groups[0][0], groups[1][0]
        s1, s2 = scores
        # cumpute new ratings
        new_r1 = self.update_rating(r1, r2, s1)
        new_r2 = self.update_rating(r2, r1, s2)
        return [[new_r1], [new_r2]]

    def expectedScore(self, rating1, rating2):
        return 1.0 / (1.0 + math.pow(10, (rating2-rating1)/self.LC))
    
    def update_rating(self, rating1: float, rating2:float, score:float):
        expected_result = self.expectedScore(rating1, rating2)
        return rating1 + (self.K * (score-expected_result))


class Glicko:
    def __init__(self, minRD: float = 30.0,
                maxRD: float = 350.0,
                c: float = 63.2,
                q: float = math.log(10, math.e)/400,
                lc: int = 400):
        
        # model constant
        self.__maxRD = maxRD # maximal value of RD
        self.__minRD = minRD # minimal value of RD
        self.lc = lc # constant in function E
        self.C = c # constant used for 'inactivity decay'
        self.Q = q # no idea how to interpret this value
        
    def G(self, rd: float):
        return 1 / math.sqrt( 1 + 3*self.Q*self.Q*(rd*rd)/(math.pi*math.pi))
    
    def expectedScore(self, rating1, rating2, mode='update'):
        RDi = 0 if mode == 'update' else rating1.sigma
        RDj = rating2.sigma
        ri, rj = rating1.mu, rating2.mu
        return 1 / (1 + math.pow(10, -self.G(math.sqrt(RDi*RDi + RDj*RDj)) * (ri-rj)/400))

    def d2(self, rating1, games: List[Tuple[Any, float]]):
        ''' NOTE:
        rating have mu sigma attributes, 
        games are dict of ratings:score,
        '''
        all_EJ = []
        all_GJ = []
        for rating2, score in games:
            # get needed variables
            Ej = self.expectedScore(rating1, rating2, mode='update')
            RDj = rating2.sigma
            Gj = self.G(RDj)
            
            # store vairables
            all_EJ.append(Ej)
            all_GJ.append(Gj)
        
        # big sum
        bigSum = 0
        for Gj, Ej, in zip(all_GJ, all_EJ):
            bigSum += Gj*Gj*Ej*(1-Ej)
            
        try:
            # d2 formula 
            return 1 / (self.Q*self.Q*bigSum)
        except ZeroDivisionError:
            
            # !!! BUG: ZeroDivisionError observed with extreme rating differences
            # !!! this will now print variable of interest
            # !!! but code will run assuming maximal and mininal expected value possible between 0 and 1
            # TODO: use warning instead of print
            print('----------------------------------------------')
            print(f'Glicko d2 ERROR: {rating1}, {games}')
            print(f'{bigSum}, {all_EJ}, {all_GJ}')
            print('----------------------------------------------')
            # just assume a very low 'bigSum'
            bigSum = 0.00000000001
            return 1 / (self.Q*self.Q*bigSum)

    def prePeriod_RD(self, rating):
        '''
        implement formula presented at step1 caee (b) p.3
        '''
        new_RD = math.sqrt(rating.sigma*rating.sigma + self.C*self.C)
        # check boundaries on sigma - ??? move max() elsewhere
        return max(min(new_RD, self.__maxRD), self.__minRD)
    
    def newRating(self, rating1, games: List[Tuple[Any, float]]):
        ''' NOTE:
        rating have mu sigma attributes, 
        games are dict of ratings:score,
        '''
            
        # compute term 'a'
        d2 = self.d2(rating1, games)
        a = self.Q / ((1/(rating1.sigma*rating1.sigma)) + (1/d2))
        
        # lcompute term 'b'
        b = 0
        for rating2, score in games:
            b += self.G(rating2.sigma)*(score - self.expectedScore(rating1, rating2, mode='update'))
        
        # create new rating object to avoid'side effect'
        rating = copy.copy(rating1)
        # post Period R
        rating.mu += a*b
        # post Period RD
        rating.sigma = math.sqrt(1/( (1/rating1.sigma**2) + (1/d2) ))
        
        return rating

    def rate(self, rating, ratings: List[Any], scores: List[float]):
        # formating
        games = [(r, s) for r, s in zip(ratings, scores)]
        return self.newRating(rating, games)





from .standing import Standing
from . import rating
from .observer import GameByGame, BatchGame, KeyChecker
from .inferer import Elo, Glicko, PlayerLevel
from .datamodel import KeyModel, GaussianModel
from .standard import BasicElo, BasicGlicko


from typing import List, Tuple, Any
from typeguard import typechecked


'''
This file deals with how Player faces each other


# ???: a decorator to raise error/warning for list of odd length
'''


# --- Shuffling --- #
def riffle_shuffle(half1: List[int], half2: List[int]) -> List[int]:
    return [half[i] for i in range(len(half1)) for half in [half1, half2]]


# --- Splitting --- #
def symetric_split(elems: List[int]) -> Tuple[List[int], List[int]]:
    h = len(elems)//2
    half1 = elems[:h]
    half2 = list(reversed(elems[-h:]))
    return half1, half2

def middle_split(elems: List[int]) -> Tuple[List[int], List[int]]:
    h = len(elems)//2
    half1 = elems[:h]
    half2 = elems[-h:]
    return half1, half2

def neighboor_split(elems: List[int]) -> Tuple[List[int], List[int]]:
    half1 = elems[::2]
    half2 = elems[1::2]
    return half1, half2

# --- matching --- #
def symetric_match(elems: List[Any]) -> List[List[Any]]:
    return [[elems[i], elems[-i]] for i in range(len(elems)//2)]

def parallel_match(elems: List[Any]) -> List[List[Any]]:
    h = len(elems)//2
    return [[elems[i], elems[h+i]] for i in range(h)]

def neighboor_match(elems: List[Any]) -> List[List[Any]]:
    return [[elems[2*i], elems[2*i+1]] for i in range(len(elems)//2)]




from typing import List

from rstt import Duel
from rstt.scheduler.tournament.competition import Competition


class Snake(Competition):
    def __init(self, *args, **kwars):
        super().__init__(*args, **kwars)
        self.snake = []

    def initialise(self):
        self.snake = [player for player in self.seeding]
        self.snake.reverse()

    def generate_games(self):
        game = Duel(self.snake.pop(0), self.snake.pop(0))
        return [game]

    def edit(self, games: List[Duel]):
        for game in games:
            # winner can will play another game
            self.snake.insert(0, game.winner())
            # loser journey ends
            self.standing[game.loser()] = len(self.snake) + 1

        # check stop condition
        if len(self.snake) == 1:
            self.standing[self.snake[0]] = 1
            finished = True
        else:
            finished = False

        return finished


from typing import Union, List, Dict, Tuple, Callable, Any, Optional

from typeguard import typechecked
import numpy as np

from rstt import Player
import rstt.utils.utils as uu


GET_SORT = 'get'
SET_SORT = 'set'
NEVER = 'never'
ALWAYS = 'always'


def get_sort(func: Callable[..., Any]) -> Callable[..., Any]:
    '''Sorting decorator
    
    This function sort the Standing before its access.
    '''
    def wrapper_get(self: Any, *args, **kwars) -> Any:
        if self._Standing__protocol in [GET_SORT, ALWAYS]:
            self._Standing__sort()
        return func(self, *args, **kwars)
    return wrapper_get
 
def set_sort(func: Callable[..., Any]) -> Callable[..., Any]:
    '''Sorting decorator
    
    This function sort the Standing after modification.

    FIXME:
        - 'lazy' sorting. Some set/edit operations are performing iteratively modifications.
        The sorting should only be called after the last operation, when the 'user call' is done.
        Example: add() and __add()
    '''
    def wrapper_set(self: Any, *args, **kwargs) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Standing__protocol in [SET_SORT, ALWAYS]:
            self._Standing__sort()
        return set_action
    return wrapper_set

'''
TODO: ADD
def not_sorted_error(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_check(self: Any, *args, **kwargs) -> Any:
        if not self._Standing__sorted:
            msg = f"Can not call {func.__name__} on unsorted Standing"
            raise AttributeError(msg) # ??? RuntimeError
        return func(self, *args, **kwargs)
    return wrapper_check
'''



class Standing:
    ''' NOTE:
        1. Implementation choice
            - Standing is an hybrid of container/iterable. It belongs to the Collections categorie as a list and dictionary.
            - It is a list, a defaultdict (key, value ) and a 'symetrical dictionary' (key, index) <-> (index, key)
            - A double implemntation (self._list and self._dict) makes many features easy and safer to implement.
            However the duplication of a data becomes a massif issue regarding synchronization and memory usage.
            - self.ranks as a List[Tuple[key, value]] as many down side, especialy because tuple are immutable.
            However it as a clarity upside: We have a list of key-value pairs.
            There are 2 explicit field, one for key, one for value.
            Items are contained in a list which implicitly assigned them an index.
            And the class build itself upon the list features.
            The dictionary-like interface needed more work arround.
            
        TODO:
        1. KEY type
            * Generalize Standing based on a key_type
            * Typehint should support the key_type in the method signature
            * ISSUE: 
                - key can not be an int, because it would have ambigous behaviour with index
                - should key_type be of different of a potential value_type
                - key_type would impact the second key for self.ranks.sort() in c.f. __sort()
            
            A naive implementation based on self.key_type will not work
            as self.key_type is not defined in a function signature
            
            Try with Generic[key_type] or Generic[key_type, value_type]
            which one should be comparable ? only value or both ?
            
            Depending on the implement choice __setitem__() can have symetrical behaviour with __getitem__()
        
        2. Sorting Policy
            * in __init__() define a sorting self.policy with
            functools.cmp_to_key : https://docs.python.org/3/library/functools.html#functools.cmp_to_key
        
        3. Typing Hints
            * decorate with @typechecked where needed
            * but without overusing it - ONLY where needed
            
        4. only_sorted_method 
            * some getter and setter do not make any sense when self.__sorted is False
            * a decorator raising error/warnings based on __sorted/_keep_sorted 
            
            for example pop() makes no sense when not sorted, remove() should be used in that case. 
            
        5. Handle __sort() calls
            * Optimizing the class performance is crucial
            * python list.sort() is optimized but remains costy on large list
            * Ideas:
                - call on either setter/getter only
                - optimal sorting algorithm explicit call based on self.ranks methods impact (local/global changes)
                - self._safety_ in ['set', 'get', 'all']
            
            Goal is to minimize calls that guarantee the expected behaviour.
            It also needs clear Warnings and Exception raised based (cf 3)
            But the optimized approach depends on the use case:
                - Is the Standing use to hold frequently updated values ?
                - Is it used to order automaticly items ?
            
            The _safety_ approach moves part of the issue to the responsability of the user
            thus add flexibility to class 'reasonable' applications.
                  
        6. full documentation
            * for user
            * for develloper
            
        7. Classmethod
            * rerank() and fit() could be done as classmethod
            * idealy they are implemented such that any inherited class work as intended
            
        8. Terminology
            * Consistency in wording for parameters and documentation
            * do we talk about keys or player, do we use index or rank ?
            
        9. isinstance()
            * extremly used in __{set/del/get}item__ methods
            * is it necessary, is it the best practice ?
            
            The reason is the list & dict behaviour implement different interfaces based on types      
        '''
    
    def __init__(self, default: float = 0.0, 
                 lower: float = np.iinfo(np.int32).min,
                 upper: float = np.iinfo(np.int32).max,
                 step: float = 1.0,
                 protocol: str = SET_SORT):
        """A Standing object simulate a value-based ordered dictionary.
        
        The Standing also act as a list and the following relationship exist:
        * Index -> Key 
        * Key -> Value.
        This means that the class implementation works similary to <List> and <Dict>.
        
        The key ordering is maintain by internal mechanism as long as the Standing is manipulate with
        the provided methods and functionnalities.
        
        Examples
        --------
        
        Notes
        -----
        Ordering is computationaly demanding, which means that repeated operations on a large Standing could result
        in performances issues. It is possible to turn off the ordering features.
        This break the whole point of the class. Do it only if needed and if you know what your are doing
        
        In its current form Key are intended to be of Type <Player> and Value of type <float>.
        This may change in the future to support a more general case such as:
        Key of type <Hasable> -> Value of type <Comarapble>.

        Parameters
        ----------
        default : float, optional
            A key added to the standing without an associated value will be assign this default value, by default 0.0
        min : float, optional
            The minimal boundary for values, by default np.iinfo(np.int32).min
        max : float, optional
            The maximal boundary for values, by default np.iinfo(np.int32).max
        step : float, optional
            An interval used for insertion operation , by default 1.0
            NOTE: not completely supported, could disapear in futur version.
        """
        
        # data 
        self.ranks = [] # [(key: Player, value: float)]
        self.__default = default # default value
        
        # sorting controls
        self.__sorted = True
        self.__maintain = True
        self.__protocol = protocol # used by decorators
        
        
        # NOTE:  Untested feature. Delete ?
        self.__min = lower
        self.__max = upper
        self.__step = step

    # --- sorting algorithm --- #
    def __sort(self):
        ''' Sorting method
        
        Called by decorators when needed and sort self.ranks [(key, value)] based on values
        NOTE: (conceptual)
            - __sort is never called 'as-it', _Standing__sort is called by get/set_sort decorator
        '''
        if self.__maintain and not self.__sorted:
            # NOTE: the general idea is: (1) sort on value, (2) on key
            # TODO: use functool lib to define the 'cmp' policy
            self.ranks.sort(key=lambda x: (x[1], x[0].name()), reverse=True)
            self.__sorted = True
    
    # --- general purpose methords --- #
    @get_sort
    def plot(self, standing_name: Optional[str]="Standing"):
        """ Plot method

        Print the Standing in a text form as Index-Key-Value.
        """
        print(f"----------- {standing_name} -----------")
        for i, (k, v) in enumerate(self.ranks):
            print("{:>5} {:>20} {:>10}".format('%d.' % i, '%s' % k, '%d' % v))
    
    @typechecked
    def fit(self, keys: List[Player]):
        """Create a new Standing instance containing only the given Keys

        The Keys will have the same value as in the current Standing if present
        otherwise the default value will be assigned. 

        Parameters
        ----------
        keys : List[Player]
            list of <Player> to be present in the new Standing

        Returns
        -------
        Standing
        """
        new_standing = Standing(self.__default, self.__min, self.__max, self.__step)
        
        # add key with appropriate value
        for key in keys:
            try:
                new_standing.__add(key, self.value(key))
            except IndexError:
                new_standing.__add(key, self.__default)
        return new_standing
    
    @set_sort # FIXME: does not make any sense. No set operation on the standing instance
    @typechecked
    def rerank(self, permutation: List[int], inverse: bool = False):
        """Create a new Standing with a different ordering of its keys.

        Create a new Standing where key have different value associated. 
        Key at index 'i' will have the value  of the Key' at index permutation[i].

        Parameters
        ----------
        permutation : List[int]
            a permutation
        inverse : bool, optional
            wether permutation should be interpreted directly,
            or as its inverse funtion. If inverse is True then Key at index permutation[i] 
            will have the value of Key at index i , by default False.

        Returns
        -------
        Standing
            A rearranged ordering of the keys.
        """
        
        # create new Standing instance
        new_standing = Standing(self.__default, self.__min, self.__max, self.__step)
        
        # compute key-value pairs
        if inverse:
            keys_values = [(self[p_i], self.value(i)) for i, p_i in enumerate(permutation)]
        else: 
            keys_values = [(self[i], self.value(p_i)) for i, p_i in enumerate(permutation)]
        
        keys, values = [item[0] for item in keys_values], [item[1] for item in keys_values]
        
        # add items to the new Standing
        new_standing.add(keys, values)
        return new_standing
    
    # --- getter --- #
    @get_sort
    @typechecked
    def percentile(self, elem: Dict[Tuple[int, int], float]):
        """Output the percentile of the input

        If provided a value, 

        Parameters
        ----------
        elem : Union[Player, float]
            Either a value of a key of the standing.
            If a value is provided it can be aribitrary, if it is a key it needs to be one present in the standing

        Returns
        -------
        float
            the percentile associated to the elem in the Standing
        """
        if isinstance(elem, Player):
            value = self.value(elem)
        else:
            value = elem

        return 1.0 - len([v for v in self.values() if v > value])/len(self.values())

    @get_sort
    def ties(self) -> bool:
        """Quick overview wether the Standing contains keys with equal value.

        Standing implement an 'Ordinal Ranking' which means that two key if equal values will have a different rank.
        
        Note
        ----
        Currently the assignement of disctinct rank uses the Player.name() returned value.
        This could change in future version to support customized policy.

        Returns
        -------
        bool
            If True, the Standing contains at least a two keys with equal value.
        """
        points = set([entry[1] for entry in self.ranks])
        return len(points) != len(self)
    
    # TODO: define clearly output format
    @get_sort
    def tied_items(self) -> Dict[float, List[Player]]:
        """Method to find tied items in the Standing

        Returns
        -------
        Dict[Tuple[int, int], List[Player]]
            return a dictionary where keys are a tuple indicating the ranks range and values are the tied keys (player)
        """
        ties = {}
        values = uu.multiples([entry[1] for entry in self.ranks])
        for val in values:
            ties[val] = []
            for player in self:
                if self.value(player) == val:
                    ties[val].append(player)
        for val in values:
            top = (min(self[ties[val]]), max(self[ties[val]]))
            ties[top] = ties[val]
            del ties[val]
        return ties
    
    # --- setter --- #
    @set_sort
    @typechecked
    # TODO: investigate redundancy between sorting=False and protocol='never'
    # TODO: define protocol type (enum ?)
    def set_sorting(self, sorting: bool = None, protocol: str = None):
        """"Set the ordering properties of the Standing

        Control the internal mechanism. How/When the Standing handle the data and the sorting calls. 

        Parameters
        ----------
        sorting : bool, optional
            Define wether the Standing should be sorted or not, by default None.
            - True: the Standing will sorts itself automaticly upon future method calls.
            - False: the Standing will not sorts itself.
            - None: the current behaviour is maintained.
        protocol : str, optional
            Define which sorting stategy is used, by default None
            - 'get': An unsorted Standing will be 'updated' upon a query on its state (keys(), values(), plot(), ...).
            - 'set': An unsorted Standing will be 'updated' upon a modification of its state (add(), insert(), ...)
            - 'always': An unsorted Standing will be updated at every opportunity.
            - None: the current strategy is maintained.
        
        Warning
        -------
        This method should only be used to optimize code performance when needed and by user aware of the consequences.
        Here are some advice and relevant implementation details. Use:
            - protocol='get' when a portion of code is slow and performs a lot of 'set-operations'.
            - protocol='set' when a portion of code is slow and performs a lot of 'get-operations'.
            - protocol='always' anytime performances are not an issue.
            - protocol='never' the standign stops ordering itself automaticaly.
            - sorting=False when the order of the keys in the Standing do not matter.
        """
        
        self.__maintain = sorting
        self.__protocol = protocol

    @typechecked
    @set_sort
    def insert(self, index: int, key: Player):
        """List like insertion

        insert a key at a given position in the list.
        It will automaticly compute an appropriate value to assign.

        Parameters
        ----------
        index : int
            Position in the list 'rank' of the given key 
        key : Player
            The element to insert in the Standing
        """
        self.__setitem_index_key(index, key)

    # TODO: support other signature such as list of tuples and dictionaries - typecheck
    @set_sort
    def add(self, keys: List[Player], values: List[float] = []):
        """Add key-value pairs to the Standing

        Built-in method to add items to the Standing.
        The values are optional as Standing has a default value specified.
        
        Note
        ----
        The current implementation only supports list of keys and list of values as parameters.
        Futur version could include parameters such as dict[key, value] and iterable of tuples.

        Parameters
        ----------
        keys : List[Player]
            Element to insert in the Standing
        values : List[float], optional
            Associated values of the keys, if len(values) < len(keys) the method will add as many items (key, value)
            to the standing before using the Standing default value for the remaining keys, by default []
            
        FIXME: 
            - turning off sorting seems unecessary since __add() is not a @set_sort decorated method
            - test performances & requirements without it
        """
        
        # turn off sorting for optimisation
        should_sort = self.__maintain
        self.__maintain = False
        
        # perform iteratively addition
        for i, key in enumerate(keys):
            try:
                self.__add(key, values[i])
            except IndexError:
                self.__add(key, self.__default)
                
        # restaure Standing status
        self.__maintain = should_sort

    # --- Containers standard methods --- #
    @get_sort
    def keys(self) -> List[Player]:
        """Get method for keys

        Similar to dict.keys() but it returns a list, not a view.

        Returns
        -------
        List[Player]
            A list with all the item registered in the Stadning
        """
        return [item[0] for item in self.ranks]
    
    @get_sort
    def values(self) -> List[float]:
        """Get method for values

        Similar to dict.values() but it returns a list, not view.

        Returns
        -------
        List[float]
            A list with all the values in the Standing
        """
        return [item[1] for item in self.items()]

    @get_sort
    @typechecked
    # !!! value() is a user level method. yet it is used in many places and triggers a lot of unecessary get_sort
    def value(self, key: Union[Player, int]) -> float:
        """Get value for a single key

        Return the value associated to the key in the Standing.
        The key can be a index/rank or an element of the Standing

        Parameters
        ----------
        key : Union[Player, int]
            A key to get the value, either the index or the actual element contained in the Standing

        Returns
        -------
        float
            The associated value to the key.
        """
        if isinstance(key, int):
            return self.ranks[key][1]
        elif isinstance(key, Player):
            return self.ranks[self.index(key)][1]

    @get_sort
    @typechecked
    def items(self) -> List[Tuple[Player, float]]:
        """Get method for key value pairs

        Similar to dict.items(). It returns a list of tuples (key, value).

        Returns
        -------
        List[Tuple[Player, float]]
            List of items as tuple (key, value).
        """
        return self.ranks
    
    @get_sort 
    @typechecked
    def index(self, key: Player) -> int:
        """Get index for key

        Similar to list.index(). Standing implements an 'Ordinal Ranking' upon the set of keys based on associtaed values.

        Parameters
        ----------
        key : Player
            The key to find in the Standing.

        Returns
        -------
        int
            The rank of the key in the Standing. 
        """
        return self.keys().index(key)

    @get_sort
    @typechecked
    def pop(self, key: Union[int, List[int], Player, List[Player]]) -> Union[Player, List[Player], int, List[int]]:
        """Get and remove method for element.

        Similar list.pop() and dict.pop() with some notable distinction. It deletes the corresponding items but:
        - If key refers to index(es)then it returns the keys present at those ranks.
        - If key refers to Player(s) then it and returns the corresponding index(es)/rank(s).

        Parameters
        ----------
        key : Union[int, List[int], Player, List[Player]]
            Accessor to the Elements to delete.

        Returns
        -------
        Union[Player, List[Player], int, List[int]]
            The corresping alternative keys. 
        """
        item = self.__getitem__(key)
        self.__delitem__(key)
        return item
    
    @get_sort
    @typechecked
    def remove(self, key: Union[int, List[int], Player, List[Player]]) -> Union[Player, List[Player], int, List[int]]:
        """Remove item method

        Works as Standing.pop() but without return values.

        Parameters
        ----------
        key : Union[int, List[int], Player, List[Player]]
            Accessor of items to delete
        """
        self.__delitem__(key)
    
    # --- internal mechanism --- #
    def __add(self, key: Player, value: float = None):
        '''
        method responsible of the inclusion of new item inside the Standing
        
        REQ:
            - no typechecking because it is called internaly,
            i.e. when type should already been approved
            - this is the only funtion that should use the self.__default
            i.e. self.__add(key, self.__default) should never be called
        '''
        if key not in self:
            if value is not None:
                value = min(max(value, self.__min), self.__max)
            else:
                value = self.__default
            self.ranks.append((key, value))
            self.__sorted = False
        else:
            msg = f"Attempt to add a key ({key}) already present in the Standing {self})"
            raise KeyError(msg)

    def __delitem_index(self, index: Union[int, slice, List[int]]):
        '''
        Method responsible to delete items based on ranks/index
        
        Implements one of the __delitem__ behaviour, namely a 'list like del'
        
        REQ:
            - support int, slice, list
            - no typechecking because it is called internaly
        
        FIXME:
            - raise error/warnings based on __sorted, __maintain
        '''
        if isinstance(index, int):
            del self.ranks[index]
        elif isinstance(index, slice):
            keys = self[index]
            self.__delitem_key(keys)
        else:
            # QUEST: is it better to use the same approach as for slice, than sorting the indexes ?
            for i in sorted(index, reverse=True):
                del self.ranks[i]
                
    def __delitem_key(self, key: Union[Player, List[Player]]):
        '''
        Method responsible to delete items based on keys
        
        Implements one of the __delitem__ behaviour, namely a 'dict like del'
        
        REQ:
            - support Player and List
            - no typechecking because it is called internaly
            - Is not sorting dependant.
        '''
        if isinstance(key, Player):
            self.__delitem_index(self.index(key))
        else:
            for k in key:
                # QUEST: use self.__delitem_key(k) ?
                self.__delitem_index(self.index(k))

    def __setitem_key_value(self, key: Player, value: float):
        '''
        Method responsible to set an item value based on the key
        
        Implements one of the __setitem__ behaviour, namely a 'dict like set'
        
        REQ:
            - support set operation for key
            - no typechecking because it is called internaly
            - Is not sorting dependant.
            
        '''
        if key not in self:
            self.__add(key, value)
        else:
            self.__change_key_value(key, value)

    def __setitem_index_key(self, index: int, key: Player):
        '''
        Method responsible to set an item value based on the key
        
        Implements one of the __setitem__ behaviour.
        Conceptualy is a 'dict-like set' but for (key, index) and not (key, value).
        technically works and do the job of a list.insert() by computating a suitable value for the key.
        
        REQ:
            - no typechecking because it is called internaly
            - Is sorting dependant
        
        FIXME:
            - raise/error for operation when not sorted
            
        # BUG: 
            - does not have the expected behaviour for an-already contained key. check __add()
        '''
        if index == 0:
            point = min(self.ranks[0][1] + self.__step, self.__max)
        elif index == len(self):
            point = max(self.ranks[-1][1] - self.__step, self.__min)
        else:
            point = (self.ranks[index-1][1] + self.ranks[index][1]) / 2
        self.__add(key, point)

    def __change_key_value(self, key: Player, value: float):
        '''Method responsible to change the value of a key.
        
        This Method assign to an already present key a new value in adict-like fashion
        
        NOTE:
        In Standing a key should be present at most once (like in a dictionary, unlike in a list).
        This challenges some design concept. What does it means to insert an already existing key, 
        but at a different index, or with a different value?
        
        This method aims to match the syntax 'my_Standing_insnatce[already_existing_key] = different_value'
        with its intuitive behaviour - the one of dictionaries. 
        '''
        if key in self:
            self.__delitem_key(key)
            self.__add(key, value)
        else:
            msg = f"Attempt to change the value of a key ({key}) not present in the Standing {self}"
            raise KeyError(msg)

    # --- magic methods --- #
    @get_sort
    @typechecked
    def __getitem__(self, key: Union[int, slice, List[int], Player, List[Player]]
                    ) -> Union[Player, List[Player], int, List[int]]:
        '''
        Method responsible to get an item
        
        It implements a symetrical like relation between index and key
        for a given pair (key, index) the following line should always be True.
            Standing[key] == index && Standing[index] == key 
        
        REQ:
            - typecheking because it is at the user interface layer.
            - symetrical key, index relationship
            - support int, list and slice for Index
            - support Player and list for Key
        
        ALERT: This function implements at the user level many different behaviours. 
        No Operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.
        
        # QUEST:
            - is the use of isinstance() appropriate
            - is it the good ordering of type control ?
            
        # NOTE:
            Coding can look confusing because it uses the [] syntax inside the __getitem__() method.
            The internal mechanism used here, are the one offered by python list and tuple class.
        
        '''
        if isinstance(key, slice):
            return [player[0] for player in self.ranks][key]
        elif isinstance(key, Player):
            return self.index(key)
        elif isinstance(key, int):
            return self.ranks[key][0]
        elif isinstance(key, list) and isinstance(key[0], Player):
            return [self.index(player) for player in key]
        elif isinstance(key, list) and isinstance(key[0], int):
            players = [player[0] for player in self.ranks]
            return [players[index] for index in key]
    
    @set_sort
    @typechecked
    def __setitem__(self, key: Union[int, Player], value: Union[Player, float]):
        '''
        Method responsible set an item in the Standing
        
        It offers the interfaces
            - list-like where a key can be assigned to an index.
            works like an insert() operation
            - dict-like where a value can be assigned to a key
        
        REQ:
            - typecheking because it is at the user interface layer.
            - dict & list like __setitem__
            
        QUEST: What about Standing[key] = index
        The assignement of an index to a key is ambiguous.
        Is the parameter value to be interpreted as an index or as the value of the key-value pair.
        This issue can be adressed once The typing problematic is solved. 
        
        ALERT: This function implements at the user level many different behaviours. 
        No Operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.
        
        BUG: 
            - There are major concern regarding set for key already present, see the sub method called.
            - The set index for already present key is a challenging concept in itself. It Should
            at least raise a warning
        
        '''
        if isinstance(key, int) and isinstance(value, Player):
            self.__setitem_index_key(key, value)
        elif isinstance(key, Player) and isinstance(value, float):
            self.__setitem_key_value(key, value)
        elif isinstance(key, Player) and isinstance(value, int):
            self.__setitem_index_key(value, key)
    
    @typechecked
    def __delitem__(self, elem: Union[slice, int, List[int], Player, List[Player]]):
        '''
        Method responsible to delete item
        
        REQ:
            - typecheking because it is at the user interface layer.
            - support int, list and slice for Index
            - support Player and list for Key
        
        ALERT: This function implements at the user level many different behaviours. 
        No operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.
        
        FIXME:
            - deleting by index when the Standing is not sorted should raise error.
            - When keep_sorting is false it should raise a warning at least.
        '''
        
        if isinstance(elem, Player) or (isinstance(elem, list) and isinstance(elem[0], Player)):
            self.__delitem_key(elem)
        else:
            self.__delitem_index(elem)
    
    @typechecked
    def __contains__(self, key: Player):
        return key in [item[0] for item in self.ranks]
    
    def __len__(self):
        return len(self.ranks)
    
    @get_sort
    def __iter__(self):
        ''' 
        Iterator magic method
        
        REQ:
            - provide an iterator on the contained keys
            - should iterate in 'order'
        
        NOTE: 
            - Not sure to understand exactly what is going on here.
        It is possible to iterate on Standing without, however Standing
        would not be instance of Iterable/Collection.
        This implementation seems to maintain the proper iteration behaviour
        and it make isinstance(Standing, Iterable/Collection) True - WHICH IS GOOD
        '''
        return [entry[0] for entry in self.ranks].__iter__()

    def __str__(self):
        return str(self.ranks)

    def __repr__(self):
        return f"Standing({self.ranks})"




from typeguard import typechecked
from typing import List, Dict, Union, Tuple, Any, Optional

from rstt.stypes import Inference, RatingSystem
from rstt import Match, Player
import rstt.utils.utils as uu

# QUEST: Do Observer realy need to be typechecked ?


@typechecked
def assign_ratings(datamodel: RatingSystem, ratings: Dict[Player, Any]):
    for key, rating in ratings.items():
        datamodel.set(key, rating)
        
@typechecked
def match_formating(datamodel: RatingSystem, game: Match) -> Tuple[List[List[Player]], List[List[ Any]], List[float], List[int]]:
    teams_as_players = game.teams()
    teams_as_ratings = [[datamodel.get(player) for player in team] for team in teams_as_players]
    
    if game.live():
        scores, ranks = None, None
    else:
        scores = game.scores()
        ranks = game.ranks()
    
    return teams_as_players, teams_as_ratings, scores, ranks

@typechecked
def result_formating(players: List[List[Player]], ratings: List[List[Any]]) -> Dict[Player, Any]:
    return {p: r for p,r in zip(uu.flatten(players), uu.flatten(ratings))}


class GameByGame:
    @typechecked
    def handle_observations(self, infer: Inference,
                            datamodel: RatingSystem,
                            games: Optional[Union[Match, List[Match]]]=None,
                            event: Optional[Any]=None, # FIXME: when Tournament class is added and type defined
                            *args, **kwargs):
        if event:
            if games:
                msg = f'Incompatible parameters, Only one of games or event must be passed'
                raise ValueError(msg)
            else: 
                self.handle_observations(infer=infer, datamodel=datamodel, games=event.games())
        elif isinstance(games, Match):
            self.single_game(infer, datamodel, games, *args, **kwargs)
        else:
            for game in games:
                self.single_game(infer, datamodel, game, *args, **kwargs)
    
    @typechecked
    def single_game(self, infer: Inference, datamodel: RatingSystem, game: Match, *args, **kwargs):
        players, ratings, score, ranks = match_formating(datamodel, game)
        try: 
            output = infer.rate(ratings, ranks=ranks, *args, **kwargs)
        except:
            output = infer.rate(ratings, score, *args, **kwargs)
        new_ratings = result_formating(players, output)
        assign_ratings(datamodel, new_ratings)
    

class KeyChecker:
    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs):
        new_ratings = {}
        
        for player in datamodel.keys():
            new_ratings.update(infer.rate(player, *args, **kwargs))
        
        assign_ratings(datamodel, new_ratings)

class BatchGame:
    @typechecked
    def handle_observations(self, infer: Inference,
                            datamodel: RatingSystem,
                            games: Optional[Union[Match, List[Match]]]=None,
                            event: Optional[Any]=None, # FIXME
                            *args, **kwargs):
        
        if event:
            if games:
                msg = f'Incompatible parameters, Only one of games or event must be passed'
                raise ValueError(msg)
            else: 
                self.handle_observations(infer=infer, datamodel=datamodel, games=event.games())
        else:
            games = self.to_list(games)
            players = self.involved_keys(games)
            new_ratings = {}
            for player in players:
                player_games = self.game_of_interest(player, games)
                ratings = [datamodel.get(game.opponent(player)) for game in player_games]
                scores = [self.player_score(player, game) for game in player_games]
                rating = datamodel.get(player)
                new_ratings[player] = infer.rate(rating, ratings, scores)

            assign_ratings(datamodel, new_ratings)
    
    @typechecked      
    def player_score(self, player: Player, game: Match):
        if game.winner() == player:
            return 1.0
        elif game.loser() == player:
            return 0.0
        elif game.isdraw():
            return 0.5
        else:
            msg = f'Player {player} can not be asigned a score value (1.0, 0.5, 0.0) for Game {game}'
            raise ValueError(msg)
    
    @typechecked
    def game_of_interest(self, player: Player, games: List[Match]) -> List[Match]:
        return [game for game in games if player in game]
    
    @typechecked
    def involved_keys(self, games: List[Match]) -> List[Player]:
        winners = [game.winner() for game in games]
        losers = [game.loser() for game in games]
        return list(set(winners+losers))

    @typechecked
    def to_list(self, games: Union[Match, List[Match]]) -> List[Match]:
        if isinstance(games, Match):
            games = [games]
        return games




from typing import List, Dict, Any, Union, Callable, Optional

from rstt.config import PLAYER_DIST, PLAYER_DIST_ARGS

import names

class Player():
    def __init__(self, name: Optional[str]=None, level: Optional[float]=None) -> None:
        self.__name = name if name else names.get_full_name()
        self.__level = level if level else PLAYER_DIST(**PLAYER_DIST_ARGS)
        
        self.__achievements = []
    
    # --- getter --- #
    def name(self) -> str:
        return self.__name
    
    def level(self) -> float:
        return self.__level
        
    # --- magic methods --- #
    def __repr__(self) -> str:
        return f"Player - name: {self.__name}, level: {self.__level}"
    
    def __str__(self) -> str:
        return self.__name
    
    @classmethod
    def create(cls, nb: int, name_gen: Callable[..., str]=names.get_full_name,
               name_params: Dict[str, Any]={},
               level_dist: Callable[..., float]=PLAYER_DIST,
               level_params=PLAYER_DIST_ARGS):
        return [cls(name=name_gen(**name_params), level=level_dist(**level_params)) for i in range(nb)]




import math

def bradleyterry(level1, level2):
    #https://en.wikipedia.org/wiki/Bradley–Terry_model
    return level1/(level1 + level2)

def logistic_elo(base, diff, constant):
    # https://wismuth.com/elo/calculator.html
    return 1.0 / (1.0 + math.pow(base, -diff/constant))

def normal_elo(diff):
    # TODO: add parametrization
    # https://wismuth.com/elo/calculator.html
    # https://fr.wikipedia.org/wiki/Fonction_d%27erreur_complémentaire
    return math.erfc(-diff/((2000.0/7)*math.sqrt(2)))/2



from typing import List
from typeguard import typechecked

from rstt.player import Player

class Match():
    @typechecked
    def __init__(self, teams: List[List[Player]]) -> None:
        self.__teams = teams
        self.__scores = None
        
    # --- getter --- #
    def teams(self) -> List[List[Player]]:
        return self.__teams
    
    def players(self) -> List[List[Player]]:
        return [player for team in self.__teams for player in team]

    def opponents(self, player: Player) -> List[Player]:
        return [p for p in self.players() if p not in self.teammates(player)]
    
    def teammates(self, player: Player) -> List[Player]:
        for team in self.players():
            if player in team:
                return [p for p in team if p != player]
    
    def scores(self) -> List[float]:
        return self.__scores
    
    def ranks(self) -> List[int]:
        return [len([other for other in self.__scores if other > value]) + 1 for value in self.__scores]
    
    # ??? Game.unpack() -> dict{players, scores, ranks, weight}
    
    # --- user interface --- #
    def live(self) -> bool:
        return True if self.__scores is None else False
    
    # --- internal mechanism --- #
    def __set_result(self, result: List[float]):
        if self.__scores is not None:
            msg = f'Attempt to assign a score to a game that has already one {self}'
            raise AttributeError(msg) # ??? RuntimeError
        else:
            if not isinstance(result, list):
                msg  = f"result must be instance of List[float], received {type(result)}"
                raise ValueError(msg)
            if not isinstance(result[0], float):
                msg = f'result must be instance of List[float], received List[{type(result[0])}]'
                raise ValueError(msg)
            # TODO: check value, match lenght with teams etc.
            if len(result) != len(self._Match__teams):
                msg = f"""result lenght does not match number of teams,
                        len(result) == {len(result)}, excepted: {len(self._Match__teams)}"""
                raise ValueError(msg)
            self.__scores = result
        
    # --- magic methods --- #
    def __repr__(self) -> str:
        return str(self)
    
    def __str__(self) -> str:
        return f"{type(self)} - teams: {self.__teams}, scores: {self.__scores}"
         
    def __contains__(self, player: Player) -> bool:
        return player in self.players()
    
    
class Duel(Match):
    def __init__(self, player1: Player, player2: Player) -> None:
        super().__init__(teams=[[player1], [player2]])
        
    # --- getter --- #
    def player1(self) -> Player:
        return self._Match__teams[0][0]
    
    def player2(self) -> Player:
        return self._Match__teams[1][0]
    
    def opponent(self, player: Player):
        players = set(self.players())
        players.remove(player) # this can raise a KeyError, which is what we want
        return list(players)[0]
    
    def winner(self) -> Player:
        if not self._Match__scores:
            return None
        if self._Match__scores[0] > self._Match__scores[1]:
            return self._Match__teams[0][0]
        elif self._Match__scores[0] < self._Match__scores[1]:
            return self._Match__teams[1][0]
        else:
            return None
        #return self._Match__teams[0][0] if self._Match__scores[0] > self._Match__scores[1] else self._Match__teams[1][0]
    
    def loser(self) -> Player:
        if not self._Match__scores:
            return None
        if self._Match__scores[0] > self._Match__scores[1]:
            return self._Match__teams[1][0]
        elif self._Match__scores[0] < self._Match__scores[1]:
            return self._Match__teams[0][0]
        else:
            return None
        #return self._Match__teams[0][0] if self._Match__scores[1] > self._Match__scores[0] else self._Match__teams[1][0]
    
    def isdraw(self) -> bool:
        if not self._Match__scores:
            return False
        return True if self._Match__scores[0] == self._Match__scores[1] else False    